/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of aug_eos5. The full aug_eos5 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * aug_eos5 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "HE5_HdfEosDef.h"
#define _C_VECTOR_IMPL
#include "util.h"


void init_namelist_t(namelist_t *e)
{
  challoc_init(&e->name);
}

void free_namelist_t(namelist_t *e)
{
  challoc_free(&e->name);
}

/** 
 * @brief a dimension in EOS2 is a pair of dimension name and size
 */
void init_dimension_t(dimension_t *e)
{
  challoc_init(&e->name);
  challoc_init(&e->max);
}

void free_dimension_t(dimension_t *e)
{
  challoc_free(&e->name);
  challoc_free(&e->max);
}

/** 
 * @brief EOS2 Swath Datafield and Geofield are mapped to this structure
 */
void init_fieldinfo_t(fieldinfo_t *e)
{
  challoc_init(&e->name);
  e->type = -1;
  dimensionalloc_init(&e->dims);
  challoc_init(&e->data);
  challoc_init(&e->filler);
  attributealloc_init(&e->localattr);
}

void free_fieldinfo_t(fieldinfo_t *e)
{
  challoc_free(&e->name);
  if (e->type != -1)
    H5Tclose(e->type);
  dimensionalloc_free(&e->dims);
  challoc_free(&e->data);
  challoc_free(&e->filler);
  attributealloc_free(&e->localattr);
}

/** 
 * @brief EOS5 attributes
 */
void init_attribute_t(attribute_t *e)
{
  challoc_init(&e->name);
  e->type = -1;
  challoc_init(&e->value);
}

void free_attribute_t(attribute_t *e)
{
  challoc_free(&e->name);
  if (e->type != -1)
    H5Tclose(e->type);
  challoc_free(&e->value);
}

/** 
 * @brief Grid information
 */
void init_gridinfo_t(gridinfo_t *e)
{
}

void free_gridinfo_t(gridinfo_t *e)
{
}

/** 
 * @brief Grid projection information
 */
void init_projinfo_t(projinfo_t *e)
{
}

void free_projinfo_t(projinfo_t *e)
{
}

/** 
 * @brief one big data structure that holds everything in one EOS5 Grid
 */
void init_grid_t(grid_t *e)
{
  init_gridinfo_t(&e->info);
  init_projinfo_t(&e->proj);
  dimensionalloc_init(&e->dim);
  fieldinfoalloc_init(&e->field);
  attributealloc_init(&e->attr);
  attributealloc_init(&e->grpattr);
}

void free_grid_t(grid_t *e)
{
  free_gridinfo_t(&e->info);
  free_projinfo_t(&e->proj);
  dimensionalloc_free(&e->dim);
  fieldinfoalloc_free(&e->field);
  attributealloc_free(&e->attr);
  attributealloc_free(&e->grpattr);
}

void init_dimmap_t(dimmap_t *e)
{
  challoc_init(&e->data);
  challoc_init(&e->geo);
}

void free_dimmap_t(dimmap_t *e)
{
  challoc_free(&e->data);
  challoc_free(&e->geo);
}

void init_indexmap_t(indexmap_t *e)
{
  challoc_init(&e->geo);
  challoc_init(&e->data);
  longalloc_init(&e->indices);
}

void free_indexmap_t(indexmap_t *e)
{
  challoc_free(&e->geo);
  challoc_free(&e->data);
  longalloc_free(&e->indices);
}

void init_swath_t(swath_zonal_t *e)
{
  dimensionalloc_init(&e->dim);
  dimmapalloc_init(&e->dimmap);
  indexmapalloc_init(&e->index);
  fieldinfoalloc_init(&e->geofield);
  fieldinfoalloc_init(&e->datafield);
  attributealloc_init(&e->attr);
  attributealloc_init(&e->grpattr);
}

void free_swath_t(swath_zonal_t *e)
{
  dimensionalloc_free(&e->dim);
  dimmapalloc_free(&e->dimmap);
  indexmapalloc_free(&e->index);
  fieldinfoalloc_free(&e->geofield);
  fieldinfoalloc_free(&e->datafield);
  attributealloc_free(&e->attr);
  attributealloc_free(&e->grpattr);
}

/** 
 * @brief parse a string with given separator
 * 
 * @param s source string
 * @param len length of @s
 * @param sep separator
 * @param[out] names vector of parsed lists
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int parse_list(const char *s, size_t len, char sep, namelistalloc *names)
{
  size_t i, j;
  int failinside = 0;

  for (i = j = 0; j <= len; ++j) {
    if ((j == len && len) || s[j] == sep) {
      namelist_t *elem;
      if (!namelistalloc_readyplus(names, 1)) { failinside = 1; earlynomem_break() }
      elem = &names->s[names->len++];
      init_namelist_t(elem);
      {
	if (!challoc_copyb(&elem->name, s + i, j - i)) { failinside = 1; earlynomem_break() }
      }
      i = j + 1;
      continue;
    }
  }
  return failinside ? FAIL : SUCCEED;
}


DECLARE_VECTOR_ALLOCATOR(contextalloc, context_t *)
contextalloc contexts;

static void init_context()
{
  static int contexts_init = 0;
  if (!contexts_init) {
    contextalloc_init(&contexts);

    contexts_init = 1;
  }
}

hid_t create_context(int augmentmethod)
{
  context_t *newctx = NULL;
  hid_t ret = -1;
  int succeeded = 0;

  do {
    init_context();

    if ((newctx = malloc(sizeof(context_t))) == NULL) earlyerr1_break(3, "ctx")
    if (!contextalloc_readyplus(&contexts, 1)) earlyerr1_break(3, "ctxid list")
    contexts.s[contexts.len] = newctx;
    ret = (hid_t)contexts.len;
    contexts.len++;

    {
      newctx->infname = NULL;
      newctx->gridfd = -1;
      namelistalloc_init(&newctx->gridlist);
      newctx->swathfd = -1;
      namelistalloc_init(&newctx->swathlist);
      attributealloc_init(&newctx->globalattr);

      newctx->hdf5fd = -1;
      newctx->augmentmethod = augmentmethod;

      dtorfncalloc_init(&newctx->dtors);

      newctx->dimnames = NULL;
      newctx->geonames = NULL;
      newctx->dimgeo_len = 0;

      newctx->dim_data_name = NULL;
      newctx->dim_data = NULL;
      newctx->dim_data_len = 0;

      newctx->dim_index_name = NULL;
      newctx->dim_index_len = 0;

      newctx->grid_swath_name = NULL;
      newctx->ptr_field = NULL;
      newctx->ptr_file = NULL;
      newctx->ptr_index = NULL;
      newctx->grid_swath_len = 0;

      newctx->instrument_name = NULL;
    }

    succeeded = 1;
  } while (0);

  if (!succeeded) {
    if (newctx != NULL) {
      namelistalloc_free(&newctx->gridlist);
      namelistalloc_free(&newctx->swathlist);
      free(newctx);
    }
  }
  return succeeded ? ret : -1;
}

void close_context(hid_t ctxid)
{
  size_t i;
  context_t *ctx = get_context(ctxid);
  if (!ctx) return;

  contexts.s[(size_t)ctxid] = NULL;

  for (i = 0; i < ctx->dtors.len; ++i)
    ctx->dtors.s[ctx->dtors.len - i - 1](ctxid);

  {
    int i;

    ctx->infname = NULL;
    ctx->gridfd = -1;
    namelistalloc_free(&ctx->gridlist);
    ctx->swathfd = -1;
    namelistalloc_free(&ctx->swathlist);
    attributealloc_free(&ctx->globalattr);

    ctx->hdf5fd = -1;

    dtorfncalloc_free(&ctx->dtors);

    if(ctx->dimnames && ctx->geonames)
    {
      for(i=0; i<ctx->dimgeo_len; i++)
      {
	free(ctx->dimnames[i]);
	free(ctx->geonames[i]);
      }
    }
    free(ctx->dimnames);
    free(ctx->geonames);

    if(ctx->dim_data_name && ctx->dim_data)
    {
      for(i=0; i<ctx->dim_data_len; i++) {
	doublealloc_free(&(ctx->dim_data[i]));
	free(ctx->dim_data_name[i]);
      }
      free(ctx->dim_data_name);
      free(ctx->dim_data);
    }

    if(ctx->dim_index_name) 
    {
      for(i=0; i<ctx->dim_index_len; i++)
	free(ctx->dim_index_name[i]);
      free(ctx->dim_index_name);
    }

    if(ctx->grid_swath_len > 0) 
    {
      for(i=0; i<ctx->grid_swath_len; i++)
	free(ctx->grid_swath_name[i]);
      free(ctx->grid_swath_name);
      free(ctx->ptr_field);
      free(ctx->ptr_file);
      free(ctx->ptr_index);
    }

    free(ctx->instrument_name);
  }
}

int add_dtor(hid_t ctxid, dtorfnc d)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);
    if (!ctx) err1_break(3, "unknown context id")
    if (!dtorfncalloc_readyplus(&ctx->dtors, 1)) nomem_break()
    ctx->dtors.s[ctx->dtors.len] = d;
    ctx->dtors.len++;

    succeeded = 1;
  } while (0);
  return succeeded ? 0 : -1;
}

context_t * get_context(hid_t id)
{
  if (id < 0 || (size_t)id >= contexts.len)
    return 0;
  return contexts.s[(size_t)id];
}

void _suicide(const char *fname, int line)
{
  fprintf(stderr, "suicide at %s %d\n", fname, line);
  exit(1);
}

void H4toH5error_set(context_t *ctx, int level, const char *msg, const char *fname, int line)
{
  fprintf(stderr, "error(%d) at %s:%d | %s\n", level, fname, line, msg);
}

void _earlynomem1(const char *fname, int line, const char *msg)
{
  fprintf(stderr, "nomem at %s:%d | %s\n", fname, line, msg);
}

void _nomem1(hid_t ctxid, const char *fname, int line, const char *msg)
{
  context_t *ctx = get_context(ctxid);
  H4toH5error_set(ctx, 3, msg, fname, line);
}

void _nomem2(hid_t ctxid, const char *fname, int line, const char *msg, const char *a1)
{
  context_t *ctx = get_context(ctxid);
  char *s = malloc(strlen(msg) + 1 + strlen(a1) + 1);
  if (s != NULL) {
    strcpy(s, msg);
    strcat(s, " ");
    strcat(s, a1);
    H4toH5error_set(ctx, 3, s, fname, line);
    free(s);
  }
  else {
    H4toH5error_set(ctx, 3, msg, fname, line);
  }
}

void _earlyerr1(int code, const char *fname, int line, const char *msg)
{
  fprintf(stderr, "error(%d) at %s:%d | %s\n", code, fname, line, msg);
}

void _warn1(hid_t ctxid, int code, const char *fname, int line, const char *msg)
{
  fprintf(stderr, "warning(%d) at %s:%d | %s\n", code, fname, line, msg);
}

void _err1(hid_t ctxid, int code, const char *fname, int line, const char *msg)
{
  context_t *ctx = get_context(ctxid);
  H4toH5error_set(ctx, code, msg, fname, line);
}

void _err2(hid_t ctxid, int code, const char *fname, int line, const char *msg, const char *a1)
{
  context_t *ctx = get_context(ctxid);
  char *s = malloc(strlen(msg) + 1 + strlen(a1) + 1);
  if (s != NULL) {
    strcpy(s, msg);
    strcat(s, " ");
    strcat(s, a1);
    H4toH5error_set(ctx, code, s, fname, line);
    free(s);
  }
  else {
    H4toH5error_set(ctx, code, msg, fname, line);
  }
}

void _err3(hid_t ctxid, int code, const char *fname, int line, const char *msg, const char *a1, const char *a2)
{
  context_t *ctx = get_context(ctxid);
  char *s = malloc(strlen(msg) + 1 + strlen(a1) + 1 + strlen(a2) + 1);
  if (s != NULL) {
    strcpy(s, msg);
    strcat(s, " ");
    strcat(s, a1);
    strcat(s, " ");
    strcat(s, a2);
    H4toH5error_set(ctx, code, s, fname, line);
    free(s);
  }
  else {
    H4toH5error_set(ctx, code, msg, fname, line);
  }
}

void _assertfailure(const char *fname, int line)
{
  fprintf(stderr, "assertion failure at %s %d\n", fname, line);
  exit(1);
}

static int indentlevel = 0;

static void printindent()
{
  int i;
  for (i = 0; i < indentlevel; ++i)
    printf("   ");
}

void _msg_push(hid_t ctxid, const char *s)
{
  printindent();
  printf("%s {{{\n", s);
  ++indentlevel;
}

void _msg_pop(hid_t ctxid)
{
  --indentlevel;
  printindent();
  printf("}}}\n");
}

void _msg1(hid_t ctxid, const char *s)
{
  printindent();
  printf("%s\n", s);
}

void _msg2(hid_t ctxid, const char *s1, const char *s2)
{
  printindent();
  printf("%s %s\n", s1, s2);
}

void _msg3(hid_t ctxid, const char *s1, const char *s2, const char *s3)
{
  printindent();
  printf("%s %s %s\n", s1, s2, s3);
}

void _msg4(hid_t ctxid, const char *s1, const char *s2, const char *s3, const char *s4)
{
  printindent();
  printf("%s %s %s %s\n", s1, s2, s3, s4);
}

/** 
 * @brief detect whether [YDim][XDim] or [XDim][YDim]
 * If any inconsistency is detected or none of data fields refers both XDim and YDim,
 * it is impossible to detect which dimension is major; so, this function returns an error.
 * 
 * @param grid grid information
 * @param[out] ydimmajor 1 if [YDim][XDim], 0 if [XDim][YDim]
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int detect_majordim(const grid_t *grid, int *ydimmajor)
{
  hsize_t i, j;
  int ym = -1;

  for (i = 0; i < grid->field.len; ++i) {
    const fieldinfo_t *field = &grid->field.s[i];
    hsize_t xdimindex, ydimindex;
    int xdimfound = 0, ydimfound = 0, major;

    for (j = 0; j < field->dims.len; ++j) {
      const char *dimname = field->dims.s[j].name.s;
      if (strcmp(dimname, "XDim") == 0) {
	xdimindex = j;
	xdimfound = 1;
      }
      else if (strcmp(dimname, "YDim") == 0) {
	ydimindex = j;
	ydimfound = 1;
      }
    }
    if (!xdimfound || !ydimfound) continue;

    major = ydimindex < xdimindex ? 1 : 0;
    if (ym == -1) ym = major;
    else if (ym != major)
      return -1;
  }
  if (ym == -1)
    return -1;
  *ydimmajor = ym;
  return 0;
}

/** 
 *  this function is part of EOS5 library.
 */
herr_t HE5_GDij2ll(int projcode, int zonecode, double projparm[], int spherecode, long xdimsize, long ydimsize, double upleftpt[], double lowrightpt[], long npnts, long row[], long col[], double longitude[], double latitude[], int pixcen, int pixcnr);

/** 
 * @brief generate longitudes and latitudes dimension scale from the given projection info
 * This function pessimistically assumes that two-dimensional coordinate variables are needed.
 * The client can detect if lons and lats can be one-dimensional by calling check_ortho_dimscale().
 * @see check_ortho_dimscale
 * 
 * @param ctxid context identifier
 * @param grid grid information
 * @param [out] lons generated longitudes
 * @param [out] lats generated latitudes
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int generate_lonlat(hid_t ctxid, const grid_t *grid, int *ydimmajor, doublealloc *lons, doublealloc *lats)
{
  long *rows = NULL, *cols = NULL;
  gridinfo_t *g = (gridinfo_t *)&grid->info;
  projinfo_t *p = (projinfo_t *)&grid->proj;
  int succeeded = 0;

  do {
    int i, j, k;
    long numpoints = grid->info.xdim * grid->info.ydim;

    if (detect_majordim(grid, ydimmajor) == -1) err1_break(3, "cannot detect major dim")

    if ((rows = malloc(numpoints * sizeof(long))) == NULL) nomem1_break("'rows'")
    if ((cols = malloc(numpoints * sizeof(long))) == NULL) nomem1_break("'cols'")
    if (*ydimmajor == 1) {
      /* rows             cols
       *   /- xdim -\       /- xdim -\
       *   0 0 0 ... 0      0 1 2 ... x
       *   1 1 1 ... 1      0 1 2 ... x
       *       ...              ...
       *   y y y ... y      0 1 2 ... x
       */
      for (k = j = 0; j < grid->info.ydim; ++j) {
	for (i = 0; i < grid->info.xdim; ++i) {
	  rows[k] = j;
	  cols[k] = i;
	  ++k;
	}
      }
    }
    else if (*ydimmajor == 0) {
      /* rows             cols
      *   /- ydim -\       /- ydim -\
      *   0 1 2 ... y      0 0 0 ... y
      *   0 1 2 ... y      1 1 1 ... y
      *       ...              ...
      *   0 1 2 ... y      2 2 2 ... y
      */
      for (k = j = 0; j < grid->info.xdim; ++j) {
	for (i = 0; i < grid->info.ydim; ++i) {
	  rows[k] = i;
	  cols[k] = j;
	  ++k;
	}
      }
    }
    else {
      err1_break(3, "internal error")
    }

    if (!doublealloc_ready(lons, numpoints)) nomem1_break("'lons'")
    if (!doublealloc_ready(lats, numpoints)) nomem1_break("'lats'")

    if (HE5_GDij2ll(grid->proj.proj, grid->proj.zone, p->param, grid->proj.sphere, grid->info.xdim, grid->info.ydim, g->upleft, g->lowright, numpoints, rows, cols, lons->s, lats->s, grid->proj.pix, grid->proj.origin) == -1)
      err1_break(3, "GDij2ll failure")
    lons->len = numpoints;
    lats->len = numpoints;

    succeeded = 1;
  } while (0);

  if (1) {
    if (rows != NULL) free(rows);
    if (cols != NULL) free(cols);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief check if lons and lats can be one-dimensional
 * @see detect_majordim
 * @see generate_lonlat
 * 
 * @param grid grid information
 * @param ydimmajor major dimension detected by detect_majordim()
 * @param lon two-dimensional longitude values retrieved by generate_lonlat()
 * @param lat two-dimensional latitude values retrieved by generate_lonlat()
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int check_ortho_dimscale(const grid_t *grid, int ydimmajor, const doublealloc *lon, const doublealloc *lat)
{
  int i, j;

  if (ydimmajor) {
    /* check if longitude is like the following: */
    /*   /- xdim - \ */
    /*   1 2 3 ... x */
    /*   1 2 3 ... x */
    /*   1 2 3 ... x */
    /*      ... */
    /*   1 2 3 ... x */
    for (i = 0; i < grid->info.xdim; ++i) {
      double v = lon->s[i];
      for (j = 1; j < grid->info.ydim; ++j) {
	if (lon->s[j * grid->info.xdim + i] != v)
	  break;
      }
      if (j != grid->info.ydim)
	return 0;
    }

    /* check if latitude is like the following: */
    /*   /- xdim -\ */
    /*   1 1 1 ... 1 */
    /*   2 2 2 ... 2 */
    /*   3 3 3 ... 3 */
    /*      ... */
    /*   y y y ... y */
    for (i = 0; i < grid->info.ydim; ++i) {
      double v = lat->s[i * grid->info.xdim];
      for (j = 1; j < grid->info.xdim; ++j) {
	if (lat->s[i * grid->info.xdim + j] != v)
	  break;
      }
      if (j != grid->info.xdim)
	return 0;
    }
  }
  else {
    /* check if longitude is like the following: */
    /*   /- ydim -\ */
    /*   1 1 1 ... 1 */
    /*   2 2 2 ... 2 */
    /*   3 3 3 ... 3 */
    /*      ... */
    /*   x x x ... x */
    for (i = 0; i < grid->info.xdim; ++i) {
      double v = lon->s[i * grid->info.ydim];
      for (j = 1; j < grid->info.ydim; ++j) {
	if (lon->s[i * grid->info.ydim + j] != v)
	  break;
      }
      if (j != grid->info.ydim)
	return 0;
    }

    /* check if latitude is like the following: */
    /*   /- ydim - \ */
    /*   1 2 3 ... y */
    /*   1 2 3 ... y */
    /*   1 2 3 ... y */
    /*      ... */
    /*   1 2 3 ... y */

    for (i = 0; i < grid->info.ydim; ++i) {
      double v = lat->s[i];
      for (j = 1; j < grid->info.xdim; ++j) {
	if (lat->s[j * grid->info.ydim + i] != v)
	  break;
      }
      if (j != grid->info.xdim)
	return 0;
    }
  }
  return 1;
}

#define NUL '\0'

char *trim(char *str)
{
      char *ibuf = str, *obuf = str;
      int i = 0, cnt = 0;

      /*
      **  Trap NULL
      */

      if (str)
      {
            /*
            **  Remove leading spaces (from RMLEAD.C)
            */

            for (ibuf = str; *ibuf && isspace(*ibuf); ++ibuf)
                  ;
            if (str != ibuf)
                  memmove(str, ibuf, ibuf - str);

            /*
            **  Collapse embedded spaces (from LV1WS.C)
            */

            while (*ibuf)
            {
                  /*if (isspace(*ibuf) && cnt)
                        ibuf++;
                  else
                  {
                        if (!isspace(*ibuf))
                              cnt = 0;
                        else
                        {
                              *ibuf = ' ';
                              cnt = 1;
                        }
                        obuf[i++] = *ibuf++;
                  }*/
		  obuf[i++] = *ibuf++;
            }
            obuf[i] = NUL;

            /*
            **  Remove trailing spaces (from RMTRAIL.C)
            */

            while (--i >= 0)
            {
                  if (!isspace(obuf[i]))
                        break;
            }
            obuf[++i] = NUL;
      }
      return str;
}


// vim:ts=8:sw=2:sts=2:cindent
