/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of aug_eos5. The full aug_eos5 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * aug_eos5 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#ifndef DUMP_UTILITY_H
#define DUMP_UTILITY_H

#include "cvct.h"
#include "H5public.h"
#include "H5Ipublic.h"

/** 
 * @brief vector-like structures
 * the following preprocessor statements create C++ vector like structures and functions.
 * e.g. DECLARE_VECTOR_ALL(challoc, char) will make bunch of functions whose names are
 * begin with challoc_, and a structure challoc which contains 's' field for actual data and
 * 'len' field for the length.
 */
DECLARE_VECTOR_ALL(challoc, char)
DECLARE_VECTOR_ALLOCATOR(longalloc, long)
DECLARE_VECTOR_ALLOCATOR(doublealloc, double)
DECLARE_VECTOR_ALLOCATOR(hobjreftalloc, hobj_ref_t)


typedef struct {
  challoc name;
} namelist_t;
void init_namelist_t(namelist_t *e);
void free_namelist_t(namelist_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(namelistalloc, namelist_t)

/** 
 * @brief a dimension in EOS2 is a pair of dimension name and size
 */
typedef struct {
  challoc name;
  hsize_t dimsize;
  challoc max;
} dimension_t;
void init_dimension_t(dimension_t *e);
void free_dimension_t(dimension_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(dimensionalloc, dimension_t)

/** 
 * @brief EOS2 attributes
 */
typedef struct {
  challoc name;
  hid_t type;
  challoc value;
} attribute_t;
void init_attribute_t(attribute_t *e);
void free_attribute_t(attribute_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(attributealloc, attribute_t)

/** 
 * @brief EOS2 Swath Datafield and Geofield are mapped to this structure
 */
typedef struct {
  challoc name;
  int rank;
  hid_t type;
  dimensionalloc dims;
  challoc data;
  challoc filler;
  attributealloc localattr; // local attributes from HE5_GDinqlocattrs
} fieldinfo_t;
void init_fieldinfo_t(fieldinfo_t *e);
void free_fieldinfo_t(fieldinfo_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(fieldinfoalloc, fieldinfo_t)

/** 
 * @brief Grid information
 */
typedef struct {
  long xdim;
  long ydim;
  double upleft[2];
  double lowright[2];
} gridinfo_t;
void init_gridinfo_t(gridinfo_t *e);
void free_gridinfo_t(gridinfo_t *e);

/** 
 * @brief Grid projection information
 */
typedef struct {
  int proj;
  int zone;
  int sphere;
  double param[16];
  int pix;
  int origin;
} projinfo_t;
void init_projinfo_t(projinfo_t *e);
void free_projinfo_t(projinfo_t *e);

/** 
 * @brief one big data structure that holds everything in one EOS5 Grid
 */
typedef struct {
  gridinfo_t info;
  projinfo_t proj;
  dimensionalloc dim;
  fieldinfoalloc field;
  attributealloc attr; /* attributes from HE5_GDinqattrs */
  attributealloc grpattr; /* attributes from HE5_GDinq */
} grid_t;
void init_grid_t(grid_t *e);
void free_grid_t(grid_t *e);

/** 
 * @brief EOS5 dimension mapping
 */
typedef struct {
  challoc data;
  challoc geo;
  long offset;
  long increment;
} dimmap_t;
void init_dimmap_t(dimmap_t *e);
void free_dimmap_t(dimmap_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(dimmapalloc, dimmap_t)

/** 
 * @brief EOS5 index mapping
 */
typedef struct {
  challoc geo;
  challoc data;
  longalloc indices;
} indexmap_t;
void init_indexmap_t(indexmap_t *e);
void free_indexmap_t(indexmap_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(indexmapalloc, indexmap_t)

/** 
 * @brief one big data structure that holds everything in one EOS5 Swath
 */
typedef struct {
  dimensionalloc dim;
  dimmapalloc dimmap;
  indexmapalloc index;
  fieldinfoalloc geofield;
  fieldinfoalloc datafield;
  attributealloc attr; /* attributes from HE5_GDinqattrs */
  attributealloc grpattr; /* attributes from HE5_GDinq */
} swath_zonal_t; /* Zonal shares the same data structure with swath, so swath_t is renamed as swath_zonal_t. [04/11/2012 LD]. */
void init_swath_t(swath_zonal_t *e);
void free_swath_t(swath_zonal_t *e);

/** 
 * @brief
 */
typedef void (*dtorfnc)(hid_t ctxid);
DECLARE_VECTOR_ALLOCATOR(dtorfncalloc, dtorfnc)
typedef struct {
  const char *infname;
  hid_t gridfd;
  namelistalloc gridlist;
  hid_t swathfd;
  hid_t zonalfd;
  namelistalloc swathlist;
  attributealloc globalattr; /* global attributes from HE5_EHinqglbattrs */

  hid_t hdf5fd;
  int augmentmethod;

  dtorfncalloc dtors;

  /* Dimension name to GeoField name mapping */
  char **dimnames;
  char **geonames;
  int dimgeo_len;
  int dimgeo_max;

  /* Dimension name to dimension data mapping */
  char **dim_data_name;
  doublealloc *dim_data;
  int dim_data_len;
  int dim_data_max;

  /* Dimension name to index mapping */
  char **dim_index_name;
  int dim_index_len;
  int dim_index_max;

  /* Start index and end index */
  char **grid_swath_name;
  int *ptr_field;
  int *ptr_file;
  int *ptr_index;
  int grid_swath_len;
  int grid_swath_max;

  /* For grid only */
  char *instrument_name;
  /* For malloc issue on 64 bit Linux machine */
  hsize_t len_of_instrument_name;

  /* Zonal index */
  int za_index;
  int za;
} context_t;

hid_t create_context();
context_t * get_context(hid_t ctxid);
int add_dtor(hid_t ctxid, dtorfnc d);
void close_context(hid_t ctxid);

int parse_list(const char *s, size_t len, char sep, namelistalloc *names);

char * trim(char *str);

/* error handling */

#define suicide() _suicide(__FILE__, __LINE__)
#define earlynomem_break() { _earlynomem1(__FILE__, __LINE__, ""); break; }
#define earlyerr1_break(l, msg) { _earlyerr1(l, __FILE__, __LINE__, msg); break; }
#define earlyerr1_flagbreak(l, var, msg) { _earlyerr1(l, __FILE__, __LINE__, msg); var = 1; break; }
#define nomem_break() { _nomem1(ctxid, __FILE__, __LINE__, ""); break; }
#define nomem_flagbreak(var) { _nomem1(ctxid, __FILE__, __LINE__, ""); var = 1; break; }
#define nomem1_break(msg) { _nomem1(ctxid, __FILE__, __LINE__, msg); break; }
#define nomem1_flagbreak(var, msg) { _nomem1(ctxid, __FILE__, __LINE__, msg); var = 1; break; }
#define nomem2_break(msg, a1) { _nomem2(ctxid, __FILE__, __LINE__, msg, a1); break; }
#define nomem2_flagbreak(var, msg, a1) { _nomem2(ctxid, __FILE__, __LINE__, msg, a1); var = 1; break; }
#define err_break(l) { _err1(ctxid, l, __FILE__, __LINE__, ""); break; }
#define err_flagbreak(l, var) { _err1(ctxid, l, __FILE__, __LINE__, ""); var = 1; break; }
#define err_return(l) { _err1(ctxid, l, __FILE__, __LINE__, ""); return FAIL; }
#define warn1(l, msg) { _warn1(ctxid, l, __FILE__, __LINE__, msg); }
#define err1_break(l, msg) { _err1(ctxid, l, __FILE__, __LINE__, msg); break; }
#define err1_flagbreak(l, var, msg) { _err1(ctxid, l, __FILE__, __LINE__, msg); var = 1; break; }
#define err1_return(l, msg) { _err1(ctxid, l, __FILE__, __LINE__, msg); return FAIL; }
#define err2_break(l, msg, a1) { _err2(ctxid, l, __FILE__, __LINE__, msg, a1); break; }
#define err2_flagbreak(l, var, msg, a1) { _err2(ctxid, l, __FILE__, __LINE__, msg, a1); var = 1; break; }
#define err3_break(l, msg, a1, a2) { _err3(ctxid, l, __FILE__, __LINE__, msg, a1, a2); break; }
#define err3_flagbreak(l, var, msg, a1, a2) { _err3(ctxid, l, __FILE__, __LINE__, msg, a1, a2); var = 1; break; }
#define ASSERT(e) { if (!(e)) _assertfailure(__FILE__, __LINE__); }
void _suicide(const char *fname, int line);
void _nomem1(hid_t ctxid, const char *fname, int line, const char *msg);
void _nomem2(hid_t ctxid, const char *fname, int line, const char *msg, const char *a1);
void _earlynomem1(const char *fname, int line, const char *msg);
void _earlyerr1(int code, const char *fname, int line, const char *msg);
void _warn1(hid_t ctxid, int code, const char *fname, int line, const char *msg);
void _err1(hid_t ctxid, int code, const char *fname, int line, const char *msg);
void _err2(hid_t ctxid, int code, const char *fname, int line, const char *msg, const char *a1);
void _err3(hid_t ctxid, int code, const char *fname, int line, const char *msg, const char *a1, const char *a2);
void _assertfailure(const char *fname, int line);

/* message */
#define msg_push(s) _msg_push(ctxid, s)
#define msg_pop() _msg_pop(ctxid)
#define msg1(s) _msg1(ctxid, s)
#define msg2(s1, s2) _msg2(ctxid, s1, s2)
#define msg3(s1, s2, s3) _msg3(ctxid, s1, s2, s3)
#define msg4(s1, s2, s3, s4) _msg4(ctxid, s1, s2, s3, s4)
void _msg_push(hid_t ctxid, const char *s);
void _msg_pop(hid_t ctxid);
void _msg1(hid_t ctxid, const char *s);
void _msg2(hid_t ctxid, const char *s1, const char *s2);
void _msg3(hid_t ctxid, const char *s1, const char *s2, const char *s3);
void _msg4(hid_t ctxid, const char *s1, const char *s2, const char *s3, const char *s4);

/* conversion */
int generate_lonlat(hid_t ctxid, const grid_t *grid, int *ydimmajor, doublealloc *lons, doublealloc *lats);
int check_ortho_dimscale(const grid_t *grid, int ydimmajor, const doublealloc *lon, const doublealloc *lat);

#endif

// vim:ts=8:sw=2:sts=2:cindent
