/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of aug_eos5. The full aug_eos5 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * aug_eos5 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "hdf5.h"
#include "util.h"
#include "write.h"

extern int g_quiet;
extern int option;
extern int start_ptr_field, end_ptr_field, start_ptr_file, end_ptr_file, start_ptr_index, end_ptr_index;

#define EOSGEOFIELD  "/HDFEOS/SWATHS/%s/Geolocation Fields/%s"
#define EOSGEOFIELD2 "/HDFEOS/ZAS/%s/Data Fields/%s"
#define EOSDATAFIELD "/HDFEOS/GRIDS/%s/Data Fields/%s"

int pos;

int open_hdf5(hid_t ctxid, const char *fname)
{
  hid_t hdf5fd = -1;
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  do {
    if ((hdf5fd = H5Fopen(fname, H5F_ACC_RDWR, H5P_DEFAULT)) < 0) err1_break(3, "H5Fopen fail")

    succeeded = 1;
  } while (0);

  if (succeeded) {
    ctx->hdf5fd = hdf5fd;
  }
  else {
    if (hdf5fd != -1) H5Fclose(hdf5fd);
  }
  return succeeded ? 0 : -1;
}

int close_hdf5(hid_t ctxid)
{
  context_t *ctx = get_context(ctxid);
  if (ctx != NULL) {
    if (ctx->hdf5fd != -1) H5Fclose(ctx->hdf5fd);
    ctx->hdf5fd = -1;
    return 0;
  }
  return -1;
}

static int make_grouppath(hid_t ctxid, const char *gridswathname, int isgrid, challoc *gridpath)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);

    if (isgrid) {
      if (!challoc_copys(gridpath, "/HDFEOS/GRIDS/")) nomem1_break("/")
    }
    else {
      if (!challoc_copys(gridpath, (ctx->za==0)? "/HDFEOS/SWATHS/": "/HDFEOS/ZAS/")) nomem1_break("/") 
    }
    if (!challoc_cats(gridpath, gridswathname)) nomem1_break("gridswath groupname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int make_aug_grouppath(hid_t ctxid, const char *gridswathname, int isgrid, challoc *gridpath)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);
    int failinside = 0;

    switch (ctx->augmentmethod) {
      case 1:
	if (!challoc_copys(gridpath, "/HDFEOS/")) nomem1_flagbreak(failinside, "/")
	break;
      case 2:
      case 3:
	if (!challoc_copys(gridpath, "/NETCDF4/")) nomem1_flagbreak(failinside, "/")
	break;
      default:
	ASSERT(0);
	break;
    }
    if (failinside) break;

    if (isgrid) {
      if (!challoc_cats(gridpath, "GRIDS/")) nomem1_flagbreak(failinside, "/")
    }
    else {
      if (!challoc_cats(gridpath, (ctx->za==0)? "SWATHS/": "ZAS/")) nomem1_flagbreak(failinside, "/")
    }

    if (!challoc_cats(gridpath, gridswathname)) nomem1_break("gridswathname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief make the absolute path for the given dimension name
 * For example, if gridswathname is "grid1", isgrid is non-zero and dimname is "dim1",
 * this function will put "/HDFEOS/GRIDS/grid1/dim1".
 * 
 * @param ctxid context identifier
 * @param gridswathname grid name or swath name
 * @param isgrid is this dimension part of grid?
 * @param dimname dimension name
 * @param[out] dimpath absolute dimension path
 * 
 * @return 0 if successful
 */
static int make_dimpath(hid_t ctxid, const char *gridswathname, int isgrid, const char *dimname, challoc *dimpath)
{
  int succeeded = 0;

  do {
    if (make_aug_grouppath(ctxid, gridswathname, isgrid, dimpath) == -1) err_break(3)
    if (!challoc_cats(dimpath, "/")) nomem1_break("/")
    if (!challoc_cats(dimpath, dimname)) nomem1_break("dimname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief make the absolute path for the given dimension name
 * For example, if gridswathname is "swath1", isgrid is zero and dsetname is "geofield1",
 * this function will put "/HDFEOS/SWATHS/swath1/Geolocation Fields/geofield1".
 * 
 * @param ctxid context identifier
 * @param gridswathname grid name or swath name
 * @param isgrid is this dimension part of grid?
 * @param dsetname dataset name
 * @param isdatafield is this data field or geo-location field?
 * @param[out] dsetpath absolute dataset path
 * 
 * @return 0 if successful
 */
static int make_fieldpath(hid_t ctxid, const char *gridswathname, int isgrid, const char *dsetname, int isdatafield, challoc *dsetpath)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);

    if (make_aug_grouppath(ctxid, gridswathname, isgrid, dsetpath) == -1) err_break(3)
    if (isdatafield) {
      if (!challoc_cats(dsetpath, "/Data Fields/")) nomem1_break("/")
    }
    else {
      if (!challoc_cats(dsetpath, (ctx->za==0)? "/Geolocation Fields/" : "/Data Fields/")) nomem1_break("/")
    }
    if (!challoc_cats(dsetpath, dsetname)) nomem1_break("datasetname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

static int detect_stringtype(hid_t ctxid, hid_t type)
{
  H5T_class_t c = H5Tget_class(type);
  if (c == H5T_STRING)
    return 1;
  return 0;
}

static int write_attribute(hid_t ctxid, hid_t target, const char *name, hid_t type, const void *data, int datalen)
{
  hid_t space = -1, at = -1;
  int succeeded = 0;

  do {
    size_t typesize;

    if (detect_stringtype(ctxid, type)) {
      if ((space = H5Screate(H5S_SCALAR)) == -1) err1_break(3, "space create")
    }
    else {
      hsize_t count;
      if ((typesize = H5Tget_size(type)) == 0) err1_break(3, "unknown type size")
      count = datalen / typesize;
      if ((space = H5Screate_simple(1, &count, &count)) == -1) err1_break(3, "space create")
    }
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((at = H5Acreate(target, name, type, space, H5P_DEFAULT, H5P_DEFAULT)) < 0) err1_break(3, "attribute create")
    if (H5Awrite(at, type, data) == -1) err1_break(3, "attribute write")

    succeeded = 1;
  } while (0);

  if (1) {
    if (at != -1) H5Aclose(at);
    if (space != -1) H5Sclose(space);
  }
  return succeeded ? 0 : -1;
}

static int write_attribute_string_internal(hid_t ctxid, hid_t dsetid, const char *name, const char *value, int addnull)
{
  hid_t tid = -1;
  int succeeded = 0;

  do {
    int valuelen = strlen(value);

    if ((tid = H5Tcopy(H5T_C_S1)) == -1) err1_break(3, "type create")
    if (H5Tset_size(tid, strlen(value) + (addnull ? 1 : 0)) == -1) err1_break(3, "type size")
    if (H5Tset_strpad(tid, H5T_STR_NULLTERM) == -1) err1_break(3, "type pad")

    if (write_attribute(ctxid, dsetid, name, tid, value, valuelen) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    if (tid != -1) H5Tclose(tid);
  }
  return succeeded ? 0 : -1;
}

static int write_attribute_string(hid_t ctxid, hid_t dsetid, const char *name, const char *value)
{
  return write_attribute_string_internal(ctxid, dsetid, name, value, 0);
}

static int write_attribute_string0(hid_t ctxid, hid_t dsetid, const char *name, const char *value)
{
  return write_attribute_string_internal(ctxid, dsetid, name, value, 1);
}

/** 
 * @brief create a group specified by path and ancestor groups if necessary
 * 
 * @param ctxid context identifier
 * @param path absolute path to create
 * 
 * @return 0 if successful
 */
static hid_t create_group_rec(hid_t ctxid, const challoc *path)
{
  hid_t group = -1;
  challoc partial;
  int succeeded = 0;

  challoc_init(&partial);
  do {
    context_t *ctx = get_context(ctxid);
    int i, failinside = 0;

    for (i = 1; i <= path->len; ++i) {
      if (i == path->len || path->s[i] == '/') {
	if (!challoc_copyb(&partial, path->s, i)) nomem1_flagbreak(failinside, "path")
	
	/* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
	if ((group = H5Gcreate(ctx->hdf5fd, partial.s, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0) err2_flagbreak(3, failinside, "create group", partial.s)

	if(!g_quiet)
	    msg2("create", partial.s);
      }
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (!succeeded) {
    if (group != -1) H5Gclose(group);
  }
  if (1) {
    challoc_free(&partial);
  }
  return succeeded ? group : -1;
}

/** 
 * @brief copy all elements in the source dataset, srcpath, to destpath
 * This is necessary for the third method. This function does not implement copying all elements yet.
 * 
 * @param ctxid context identifier
 * @param srcpath source dataset
 * @param destpath target path
 * 
 * @return 0 if successful
 */
static int clone_dataset(hid_t ctxid, const challoc *srcpath, const challoc *destpath)
{
  hid_t source = -1;
  hid_t dest = -1;
  hid_t tid = -1;
  hid_t spaceid = -1;
  hid_t plist = -1;
  hsize_t *chunksizes = NULL;
  void *buf = NULL;
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);
    hssize_t npoints;
    size_t elemsize;
    H5D_layout_t layout;

    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((source = H5Dopen(ctx->hdf5fd, srcpath->s, H5P_DEFAULT)) < 0) err2_break(3, "open dataset", srcpath->s)
    if ((tid = H5Dget_type(source)) < 0) err2_break(3, "get type", srcpath->s)
    if ((spaceid = H5Dget_space(source)) < 0) err2_break(3, "get space", srcpath->s)

    if ((npoints = H5Sget_simple_extent_npoints(spaceid)) < 0) err2_break(3, "get npoints", srcpath->s)
    if ((elemsize = H5Tget_size(tid)) == 0) err2_break(3, "get size", srcpath->s)
    if ((buf = malloc(npoints * elemsize)) == NULL) nomem2_break("clone", srcpath->s)
    if (H5Dread(source, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(3, "read", srcpath->s)

    if ((plist = H5Dget_create_plist(source)) < 0) err2_break(3, "plist", srcpath->s)
    if ((layout = H5Pget_layout(plist)) < 0) err2_break(3, "layout", srcpath->s)
    if (layout == H5D_CHUNKED) {
      int nfilter = H5Pget_nfilters(plist);
      if (nfilter < 0) err1_break(3, "nfilters")
      else if (nfilter == 0) {
	if (H5Pset_filter(plist, H5Z_FILTER_DEFLATE, 0, 0, NULL) < 0) err1_break(3, "chunk")
      }
    }
    /* Using HDF5 1.8 APIs. [06/26/2012 -LD] */
    if ((dest = H5Dcreate(ctx->hdf5fd, destpath->s, tid, spaceid, H5P_DEFAULT, plist, H5P_DEFAULT)) < 0) err2_break(3, "create dataset", destpath->s)
    if (H5Dwrite(dest, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(3, "write", destpath->s)

      if(!g_quiet)
	msg4("clone", srcpath->s, "to", destpath->s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (buf) free(buf);
    if (chunksizes) free(chunksizes);
    if (plist != -1) H5Pclose(plist);
    if (spaceid != -1) H5Sclose(spaceid);
    if (tid != -1) H5Tclose(tid);
    if (dest != -1) H5Dclose(dest);
    if (source != -1) H5Dclose(source);
  }
  return succeeded ? 0 : -1;
}

typedef struct {
  hid_t ctxid;
  const challoc *srcpath;
  const challoc *destpath;
} visit_data_t;

/** 
 * @brief copy or link one dataset
 * This function is a call back function called by H5Giterate()
 * 
 * @param srcgrp an HDF5 group under HDFEOS group
 * @param name dataset name or group name
 * @param data an instance of visit_data_t structure
 * 
 * @return 0 if successful
 */
static herr_t link_grid_children(hid_t srcgrp, const char *name, void *data)
{
  challoc srcfullpath;
  challoc destfullpath;
  hid_t group = -1;
  int succeeded = 0;

  challoc_init(&srcfullpath);
  challoc_init(&destfullpath);
  do {
    const visit_data_t *visitdata = data;
    hid_t ctxid = visitdata->ctxid;
    context_t *ctx = get_context(ctxid);
    H5G_stat_t objinfo;
    int failinside = 0;

    if (!challoc_copy(&srcfullpath, visitdata->srcpath)) nomem1_break("srcpath")
    if (!challoc_cats(&srcfullpath, "/")) nomem1_break("srcpath")
    if (!challoc_cats(&srcfullpath, name)) nomem1_break("srcpath")
    if (!challoc_copy(&destfullpath, visitdata->destpath)) nomem1_break("destpath")
    if (!challoc_cats(&destfullpath, "/")) nomem1_break("destpath")
    if (!challoc_cats(&destfullpath, name)) nomem1_break("destpath")


    if (H5Gget_objinfo(srcgrp, name, 0, &objinfo) < 0) err2_break(3, "get objinfo", srcfullpath.s)
    switch (objinfo.type) {
      case H5G_GROUP:
	/* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
	if ((group = H5Gcreate(ctx->hdf5fd, destfullpath.s, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0) err2_flagbreak(3, failinside, "create group", destfullpath.s)
	  if(!g_quiet)
	    msg2("create group", destfullpath.s);
	{
	  visit_data_t nested;
	  nested.ctxid = ctxid;
	  nested.srcpath = &srcfullpath;
	  nested.destpath = &destfullpath;
	  if(!g_quiet)
	    msg_push(srcfullpath.s);
	  if (H5Giterate(ctx->hdf5fd, srcfullpath.s, NULL, link_grid_children, &nested)) err2_flagbreak(3, failinside, "visit", srcfullpath.s)
	    if(!g_quiet)
	      msg_pop();
	}
	break;
      case H5G_DATASET:
	/* According to the augmentation method, this function will either create a link or clone a dataset */
	switch (ctx->augmentmethod) {
	  case 2:
	    if (H5Glink(ctx->hdf5fd, H5G_LINK_HARD, srcfullpath.s, destfullpath.s) < 0) err3_flagbreak(3, failinside, "link", srcfullpath.s, destfullpath.s)
	    msg4("link", destfullpath.s, "to", srcfullpath.s);
	    break;
	  case 3:
	    if (clone_dataset(ctxid, &srcfullpath, &destfullpath) == -1) err3_flagbreak(3, failinside, "clone", srcfullpath.s, destfullpath.s)
	    break;
	  default:
	    ASSERT(0);
	}
	break;
      default:
	err2_flagbreak(3, failinside, "unknown objtype", srcfullpath.s)
	break;
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (group != -1) H5Gclose(group);
    challoc_free(&srcfullpath);
    challoc_free(&destfullpath);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief create an HDF5 group called NETCDF4 and copy or link every decendant group of HDFEOS group under NETCDF4 group
 * 
 * @param ctxid context identifier
 * @param gridswathname grid name or swath name
 * @param isgrid is this part of grid?
 * 
 * @return 0 if successful
 */
int create_aug_group(hid_t ctxid, const char *gridswathname, int isgrid)
{
  challoc srcpath;
  challoc destpath;
  hid_t destgrp = -1;
  int succeeded = 0;

  challoc_init(&srcpath);
  challoc_init(&destpath);
  do {
    context_t *ctx = get_context(ctxid);
    visit_data_t visitdata;

    if (make_grouppath(ctxid, gridswathname, isgrid, &srcpath) == -1) err_break(3)
    if (make_aug_grouppath(ctxid, gridswathname, isgrid, &destpath) == -1) err_break(3)
    if ((destgrp = create_group_rec(ctxid, &destpath)) < 0) err2_break(3, "create group", destpath.s)

      if(!g_quiet)
	msg_push(srcpath.s);
    visitdata.ctxid = ctxid;
    visitdata.srcpath = &srcpath;
    visitdata.destpath = &destpath;
    if (H5Giterate(ctx->hdf5fd, srcpath.s, NULL, link_grid_children, &visitdata)) err2_break(3, "visit", srcpath.s)
      if(!g_quiet)
	msg_pop();

    succeeded = 1;
  } while (0);

  if (1) {
    if (destgrp != -1) H5Gclose(destgrp);
    challoc_free(&destpath);
    challoc_free(&srcpath);
  }
  return succeeded ? 0 : -1;
}

hid_t get_grid_dim_type(context_t *ctx, const char *gridname, const dimension_t *dim) { 
  hid_t tid = H5Tcopy(H5T_IEEE_F32LE); /* for default or dim not specified */
  int i;

  switch(option) {
    case 1: /* -i for 3d only */
      {
	if(!(strcmp(dim->name.s, "XDim")==0 || (strcmp(dim->name.s, "YDim")==0)))
	  tid = H5Tcopy(H5T_NATIVE_INT);
	break;
      }
    case 2: /* -f */
      {
        for(i=start_ptr_file; i<end_ptr_file; i++) {
	  if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
	     break;     
	}
	if(i < end_ptr_file) {/* use file */
	   tid = H5Tcopy(H5T_IEEE_F32LE);
	   goto TID_READY;
	}

	for(i=start_ptr_index; i<end_ptr_index; i++) {
	  if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
	    break;
	}
	if(i < end_ptr_index) {/* use index */
	  if(strcmp(dim->name.s, "XDim") == 0 || strcmp(dim->name.s, "YDim") == 0) {
	    printf("%s is not allowed to use index option.\n", dim->name.s);
	    exit(0);
	  }
	  tid = H5Tcopy(H5T_NATIVE_INT);
	  goto TID_READY;
	}

        for(i=start_ptr_field; i<end_ptr_field; i++) {
	  if(strcmp(dim->name.s, ctx->dimnames[i])==0)
	    break;
        }
        if(i < end_ptr_field) {
	  if(strcmp(dim->name.s, "XDim") == 0 || strcmp(dim->name.s, "YDim") == 0) {
	    printf("%s is not allowed to use geofield option.\n", dim->name.s);
	    exit(0);
	  }
	  const char *geoname = ctx->geonames[i];
	  char geopath[1024];
	  hid_t source = -1;
	  sprintf(geopath, (ctx->za==0)? EOSGEOFIELD: EOSGEOFIELD2,  gridname, geoname);
	  /* Using HDF8 1.8 APIs. [06/25/2012 -LD] */
	  if (!(source = H5Dopen(ctx->hdf5fd, geopath, H5P_DEFAULT)) < 0) 
	    tid = H5Dget_type(source);
          if (source != -1) H5Dclose(source);
	}
      }
  }
TID_READY:
  return tid;
}

hid_t get_swath_dim_type(context_t *ctx, const char *swathname, const dimension_t *dim)
{
  char geopath[1024];
  hid_t tid=-1, spaceid=-1, propid=-1, destid=-1, source;
  int i;

  tid= H5Tcopy(H5T_IEEE_F32LE); /* default */

  switch(option)
  {
  case 1:
    {
      tid = H5Tcopy(H5T_NATIVE_INT); /* -i */
      break;
    }
  case 2: 
    {
      int i;
      for(i=start_ptr_file; i<end_ptr_file; i++)
	if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
	  break;
      if(i < end_ptr_file) /* [dim][file] */
      { 
	tid = H5Tcopy(H5T_IEEE_F32LE);
	goto TID_OBTAINED;
      }
     
      for(i=start_ptr_index; i<end_ptr_index; i++)
      {
	if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
	  break;
      }
      if(i < end_ptr_index) /* [dim] */
      {
	tid = H5Tcopy(H5T_NATIVE_INT);
	goto TID_OBTAINED;
      }

      for(i=start_ptr_field; i<end_ptr_field; i++)
	if(strcmp(dim->name.s, ctx->dimnames[i])==0)
	  break;
      if(i < end_ptr_field) /* [dim][field] */
      {
	const char *geoname = ctx->geonames[i];
	sprintf(geopath, (ctx->za==0)? EOSGEOFIELD: EOSGEOFIELD2,  swathname, geoname);
	/* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
	if (!((source = H5Dopen(ctx->hdf5fd, geopath, H5P_DEFAULT)) < 0))
	  tid = H5Dget_type(source);
        if (source != -1) H5Dclose(source);
      }
    }
  }

TID_OBTAINED:
  return tid;
}

int write_hdf5_dimension(hid_t ctxid, const char *gridswathname, int isgrid, int ispure, const dimension_t *dim)
{
  hid_t tid = -1;
  hid_t spaceid = -1;
  hid_t propid = -1;
  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;

  challoc_init(&dimpath);
  do {
    context_t *ctx = get_context(ctxid);
    hsize_t dimsize = dim->dimsize;
    hsize_t maxdimsize = dim->dimsize;

    if (dimsize == H5S_UNLIMITED)
      dimsize = 0;

    if (make_dimpath(ctxid, gridswathname, isgrid, dim->name.s, &dimpath) == -1) err_break(3)

    if ((spaceid = H5Screate_simple(1, &dimsize, &maxdimsize)) < 0) err3_break(3, "H5Screate_simple", gridswathname, dim->name.s)

    if(isgrid)
      tid = get_grid_dim_type(ctx, gridswathname, dim);
    else
      tid = get_swath_dim_type(ctx, gridswathname, dim);

    if ((propid = H5Pcreate(H5P_DATASET_CREATE)) < 0) err1_break(3, "H5Pcreate")
    if (maxdimsize == H5S_UNLIMITED) {
      hsize_t chunksize = 1;
      if (H5Pset_chunk(propid, 1, &chunksize) < 0) err1_break(3, "H5Pset_chunk")
    }
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dsetid = H5Dcreate(ctx->hdf5fd, dimpath.s, tid, spaceid, H5P_DEFAULT, propid, H5P_DEFAULT)) < 0) err2_break(3, "H5Dcreate", dimpath.s)

    if (write_attribute_string0(ctxid, dsetid, "CLASS", "DIMENSION_SCALE") == -1) err_break(3)
    if (ispure) {
      /* The created dimension will not contain any meaningful value. */
      if (write_attribute_string0(ctxid, dsetid, "NAME", "This is a netCDF dimension but not a netCDF variable.") == -1) err_break(3)
    }

    if(!g_quiet)
    {
      if (ispure)
	msg2("create a pure dimension", dimpath.s);
      else
	msg2("create a dimensional scale", dimpath.s);
    }

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&dimpath);
    if (dsetid != -1) H5Dclose(dsetid);
    if (propid != -1) H5Pclose(propid);
    if (spaceid != -1) H5Sclose(spaceid);
    if (tid != -1) H5Tclose(tid);
  }
  return succeeded ? 0 : -1;
}

int write_hdf5_implicit_dimension(hid_t ctxid, const char *gridname, const grid_t *grid, int isxdim, int ispure)
{
  dimension_t dim;
  int succeeded = 0;

  init_dimension_t(&dim);
  do {
    if (!challoc_copys(&dim.name, isxdim ? "XDim" : "YDim")) nomem1_break("dim name")
    dim.dimsize = isxdim ? grid->info.xdim : grid->info.ydim;
    if (write_hdf5_dimension(ctxid, gridname, 1, ispure, &dim) == -1) err_break(3)

    succeeded = 1;
  } while (0);

  if (1) {
    free_dimension_t(&dim);
  }
  return succeeded ? 0 : -1;
}

static const char *HDF5_DIMENSION_LIST = "DIMENSION_LIST";

static int write_dimensionlist_attr(hid_t ctxid, hid_t h5dataset, const hobjreftalloc *dimscales)
{
  hvl_t *dimrefs = NULL;
  int deepfreecount = 0;
  hid_t spaceid = -1;
  hid_t tid = -1;
  hid_t attrid = -1;
  int succeeded = 0;

  do {
    if (dimscales->len) {
      hssize_t numelems;
      int i, failinside = 0;
      hsize_t spacedimsizes = dimscales->len;


      H5E_BEGIN_TRY {
	attrid = H5Aopen_name(h5dataset, HDF5_DIMENSION_LIST);
      } H5E_END_TRY;
      if (attrid != -1) {
	/* This routine may be called more than once for fake dimscales such as lon and lat.
	 * Check if previously written data are same as new input
	 */
	if ((spaceid = H5Aget_space(attrid)) == -1) err1_break(3, "H5Aget_space")
	if ((tid = H5Aget_type(attrid)) == -1) err1_break(3, "H5Aget_type")
	if ((numelems = H5Sget_simple_extent_npoints(spaceid)) == -1) err1_break(3, "get extent_npoints")
	ASSERT(numelems == dimscales->len)
	if ((dimrefs = malloc(dimscales->len * sizeof(hvl_t))) == NULL) nomem1_break("dim ref vl type")
	if (H5Aread(attrid, tid, dimrefs) == -1) err1_break(3, "read dimension_list")
	for (i = 0; i < dimscales->len; ++i) {
	  ASSERT(dimrefs[i].len == 1)
	  ASSERT(*(hobj_ref_t *)dimrefs[i].p == dimscales->s[i])
	}
      }
      else {
	if ((spaceid = H5Screate_simple(1, &spacedimsizes, NULL)) == -1) err1_break(3, "H5Screate_simple")
	/* I think the specification differs from the implementation. Both netCDF4 and H5DS
	 * implementation uses VLEN of hobj_ref_t as datatype, and I follow this.
	 */
	if ((tid = H5Tvlen_create(H5T_STD_REF_OBJ)) == -1) err1_break(3, "H5Tvlen_create failure")

	if ((dimrefs = malloc(dimscales->len * sizeof(hvl_t))) == NULL) nomem1_break("dim ref vl type")
	for (i = 0; i < dimscales->len; ++i) {
	  if ((dimrefs[i].p = malloc(sizeof(hobj_ref_t))) == NULL) nomem1_flagbreak(failinside, "dim ref vl[i].p")
	  ++deepfreecount;
	  *((hobj_ref_t *)dimrefs[i].p) = dimscales->s[i];
	  dimrefs[i].len = 1;
	}
	if (failinside) break;
	/* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
	if ((attrid = H5Acreate(h5dataset, HDF5_DIMENSION_LIST, tid, spaceid, H5P_DEFAULT, H5P_DEFAULT)) == -1) err1_break(3, "dataset attribute")
	if (H5Awrite(attrid, tid, dimrefs) == -1) err1_break(3, "write dataset attribute")
      }
    }
    succeeded = 1;
  } while (0);

  if (1) {
    if (spaceid != -1) H5Sclose(spaceid);
    if (tid != -1) H5Tclose(tid);
    if (attrid != -1) H5Aclose(attrid);
    if (dimrefs != NULL) {
      while (deepfreecount--)
	free(dimrefs[deepfreecount].p);
      free(dimrefs);
    }
  }
  return succeeded ? 0 : -1;
}

typedef struct {
  hobj_ref_t ref;
  int index;
} reference_backpointer_t;

static const char *HDF5_REFERENCE_LIST = "REFERENCE_LIST";

static int write_referencelist_attr(hid_t ctxid, hobj_ref_t dimscaleref, hobj_ref_t datasetref, int dimindex)
{
  hid_t dimscaleid = -1;
  hid_t spaceid = -1;
  hid_t oldspaceid = -1;
  hid_t tid = -1;
  hid_t attrid = -1;
  reference_backpointer_t *refs = NULL;
  int succeeded = 0;

  do {
    hssize_t numelems;
    hsize_t dimsize;
    context_t *ctx = get_context(ctxid);

    if ((dimscaleid = H5Rdereference(ctx->hdf5fd, H5R_OBJECT, &dimscaleref)) < 0) err1_break(3, "H5Rdereference")

    /* If "REFERENCE_LIST" is already there, one element should be appended. It seems that appending
     * is not supported. To handle this, append in the memory first, and then remove and recreate the attribute
     */ 
    H5E_BEGIN_TRY {
      attrid = H5Aopen_name(dimscaleid, HDF5_REFERENCE_LIST);
    } H5E_END_TRY;
    if (attrid != -1) {
      /* REFERENCE_LIST exists in ds dataset */
      if ((tid = H5Aget_type(attrid)) == -1) err1_break(3, "H5Aget_type")
      if ((oldspaceid = H5Aget_space(attrid)) == -1) err1_break(3, "H5Aget_space")
      if ((numelems = H5Sget_simple_extent_npoints(oldspaceid)) == -1) err1_break(3, "H5Sget_simple_extent_npoints")
      numelems++;
      if ((refs = malloc((int)numelems * sizeof(reference_backpointer_t))) == NULL) nomem1_break("reference_backpointer_t")
      if (H5Aread(attrid, tid, refs) == -1) err1_break(3, "H5Aread")
      if (H5Sclose(oldspaceid) == -1) err1_break(3, "H5Sclose")
      oldspaceid = -1;
      if (H5Aclose(attrid) == -1) err1_break(3, "H5Aclose")
      attrid = -1;
      if (H5Adelete(dimscaleid, HDF5_REFERENCE_LIST) == -1) err1_break(3, "H5Adelete")
    }
    else {
      numelems = 1;
      if ((refs = malloc((int)numelems * sizeof(reference_backpointer_t))) == NULL) nomem1_break("reference_backpointer_t 1")
    }
    refs[numelems - 1].ref = datasetref;
    refs[numelems - 1].index = dimindex;

    dimsize = numelems;
    if ((spaceid = H5Screate_simple(1, &dimsize, NULL)) == -1) err1_break(3, "H5Screate_simple")
    if (tid == -1) {
      if ((tid = H5Tcreate(H5T_COMPOUND, sizeof(reference_backpointer_t))) == -1) err1_break(3, "H5Tcreate")
      if (H5Tinsert(tid, "dataset", HOFFSET(reference_backpointer_t, ref), H5T_STD_REF_OBJ) == -1) err1_break(3, "H5Tinsert DATASET")
      if (H5Tinsert(tid, "index", HOFFSET(reference_backpointer_t, index), H5T_NATIVE_INT) == -1) err1_break(3, "H5Tinsert INDEX")
    }
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((attrid = H5Acreate(dimscaleid, HDF5_REFERENCE_LIST, tid, spaceid, H5P_DEFAULT, H5P_DEFAULT)) == -1) err1_break(3, "H5Acreate")
    if (H5Awrite(attrid, tid, refs) == -1) err1_break(3, "write dimension scale data set attribute")

    succeeded = 1;
  } while (0);

  if (1) {
    if (refs != NULL) free(refs);
    if (attrid != -1) H5Aclose(attrid);
    if (tid != -1) H5Tclose(tid);
    if (oldspaceid != -1) H5Sclose(oldspaceid);
    if (spaceid != -1) H5Sclose(spaceid);
    if (dimscaleid != -1) H5Dclose(dimscaleid);
  }
  return succeeded ? 0 : -1;
}

static int attach_dimension(hid_t ctxid, const char *dsetpath, const hobjreftalloc *dimscales)
{
  hid_t dsetid = -1;
  hobj_ref_t dsetref;
  int succeeded = 0;

  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);

    /* DIMENSION_LIST */
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dsetid = H5Dopen(ctx->hdf5fd, dsetpath, H5P_DEFAULT)) < 0) err2_break(3, "H5Dopen", dsetpath)
    if (write_dimensionlist_attr(ctxid, dsetid, dimscales) == -1) err_break(3)

    /* REFERENCE_LIST */
    if (H5Rcreate(&dsetref, ctx->hdf5fd, dsetpath, H5R_OBJECT, -1) < 0) err2_break(3, "H5Rcreate dset", dsetpath)
    for (i = 0; i < dimscales->len; ++i) {
      if (write_referencelist_attr(ctxid, dimscales->s[i], dsetref, i) == -1) err_flagbreak(3, failinside)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
  }
  return succeeded ? 0 : -1;
}

int associate_hdf5_dimension(hid_t ctxid, const char *gridswathname, int isgrid, const fieldinfo_t *field, int isdatafield)
{
  hobjreftalloc dimscales;
  challoc dimpath;
  challoc dimnames;
  challoc dsetpath;
  int succeeded = 0;

  hobjreftalloc_init(&dimscales);
  challoc_init(&dimpath);
  challoc_init(&dimnames);
  challoc_init(&dsetpath);
  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);

    if (!hobjreftalloc_ready(&dimscales, field->dims.len)) nomem1_break("obj ref")

      for (i = 0; i < field->dims.len; ++i) {
      const dimension_t *dim = &field->dims.s[i];

      if (make_dimpath(ctxid, gridswathname, isgrid, dim->name.s, &dimpath) == -1) err_flagbreak(3, failinside)
      if (!challoc_cats(&dimnames, i == 0 ? "[" : ",")) nomem1_flagbreak(failinside, "dim names")
      if (!challoc_cats(&dimnames, dim->name.s)) nomem1_flagbreak(failinside, "dim names")
      if (H5Rcreate(&dimscales.s[i], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err2_flagbreak(3, failinside, "H5Rcreate dim", dimpath.s)
      dimscales.len++;
    }

    if (failinside) break;
    if (!challoc_cats(&dimnames, "]")) nomem1_flagbreak(failinside, "dim names")

    if (make_fieldpath(ctxid, gridswathname, isgrid, field->name.s, isdatafield, &dsetpath) == -1) err_break(3)
    if (attach_dimension(ctxid, dsetpath.s, &dimscales) == -1) err_break(3)

      if(!g_quiet)
	msg4("associate dimension", dsetpath.s, "with", dimnames.s);

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&dimpath);
    challoc_free(&dimnames);
    challoc_free(&dsetpath);
    hobjreftalloc_free(&dimscales);
  }
  return succeeded ? 0 : -1;
}

static hid_t get_lonlat_type(hid_t ctxid, const char *gridname, int islongitude) {
  hid_t source = -1, tid = -1;
  char path[1024];
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;
  
  if(!islongitude)
    sprintf(path, EOSDATAFIELD, gridname, "Latitude");
  else
    sprintf(path, EOSDATAFIELD, gridname, "Longitude");
  
  do {
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if((source = H5Dopen(ctx->hdf5fd, path, H5P_DEFAULT)) < 0) err2_break(3, "H5Dopen", path)
    if((tid = H5Dget_type(source)) < 0) err2_break(3, "H5Dget_type", path)
    succeeded = 1;
  } while(0);

  if(source != -1) H5Dclose(source);

  return succeeded? tid: -1;
}

void* get_coord_from_geofield(hid_t ctxid, const char *gridname, int islongitude) {
  int succeeded = 0;
  void* coord = NULL;
  
  do {
    hid_t source = -1, tid = -1, spaceid = -1;
    hssize_t npoints;
    size_t elemsize;
    context_t* ctx = get_context(ctxid);
    char path[1024];
    if(!islongitude)
      sprintf(path, EOSDATAFIELD, gridname, "Latitude");
    else
      sprintf(path, EOSDATAFIELD, gridname, "Longitude");
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if((source = H5Dopen(ctx->hdf5fd, path, H5P_DEFAULT)) < 0) err2_break(3, "H5Dopen", path)
    if((tid = H5Dget_type(source)) < 0) err2_break(3, "H5Dget_type", path)
    if((spaceid = H5Dget_space(source)) < 0) err2_break(3, "H5Dget_space", path) 
    
    H5T_class_t  type_class = H5Tget_class(tid);

    if((npoints = H5Sget_simple_extent_npoints(spaceid)) < 0) err2_break(3, "H5Sget_simple_extent_npoints", path)  						
    if((elemsize = H5Tget_size(tid)) < 0) err2_break(3, "H5Tget_size", path)
    if((coord = malloc(npoints*elemsize)) == NULL) nomem2_break("malloc", path)
    if(H5Dread(source, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, coord) < 0) err2_break(3, "H5Dread", path)										
    if(source != -1)H5Dclose(source);
    if(tid != -1) H5Tclose(tid);

    succeeded = 1;
  } while(0); 

  return succeeded? coord: NULL;
} 

int fill_hdf5_1d_3rd_dimension(hid_t ctxid, const char *gridname, const dimension_t  *dim) {
   hid_t dsetid = -1;
   hid_t memtype = -1;
   hid_t spaceid = -1;
   hid_t atype = -1;
   challoc dimpath;
   hobjreftalloc dimscales;
   int succeeded = 0;

   float* fbuf = NULL;

   challoc_init(&dimpath);
   hobjreftalloc_init(&dimscales);

   do {
     context_t *ctx = get_context(ctxid);
     if(make_dimpath(ctxid, gridname, 1, dim->name.s, &dimpath) == -1) err_break(3)
     switch(option) {
       case 0:
	 {
	   put_att(ctxid, dim, dimpath);
	   break;
	 }
       case 1:
	 {
	   fill_dim_scale_with_index(ctxid, dim, dimpath);
	   break;
	 }
       case 2:
	 {
	   int i;
	   for(i=start_ptr_index; i<end_ptr_index; i++) {
	     if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
	        break;
	   }
	  
	   if(i < end_ptr_index) {/* use index */
	     fill_dim_scale_with_index(ctxid, dim, dimpath);
	     goto FINISHED; 
	   }

	   for(i=start_ptr_file; i<end_ptr_file; i++) {
             if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
               break;
           } 
           if(i < end_ptr_file) {/* use file */
	     pos = i;
	     fill_dim_scale_with_data(ctxid, dim, dimpath);
             goto FINISHED;
           } 

	   for(i=start_ptr_field; i<end_ptr_field; i++) {/* use data field */
	     if(strcmp(dim->name.s, ctx->dimnames[i])==0)
               break;
	   }
	   if(i < end_ptr_field) {
	     fill_dim_scale_with_field(ctxid, gridname, dim, dimpath, 1, i);
	   }
	 } /* end of case */
     } /* end of switch */
FINISHED:	         
     if (!hobjreftalloc_ready(&dimscales, 1)) nomem1_break("obj ref")
     if (H5Rcreate(&dimscales.s[0], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err1_break(3, "H5Rcreate")	
     dimscales.len++;

     if (attach_dimension(ctxid, dimpath.s, &dimscales) == -1) err_break(3)

     if(!g_quiet)
       msg2("fill 1d dimension", dimpath.s);

      succeeded = 1;
   } while (0);

  if (1) {
    hobjreftalloc_free(&dimscales);
    challoc_free(&dimpath);
  }

  return succeeded ? 0 : -1;
}

int fill_hdf5_1d_dimension(hid_t ctxid, const char *gridname, int islongitude, const doublealloc *coord)
{
  hid_t dsetid = -1;
  hid_t memtype = -1;
  hid_t spaceid = -1;
  hid_t atype = -1;
  challoc dimpath;
  hobjreftalloc dimscales;
  int succeeded = 0;

  float* fbuf = NULL;

  challoc_init(&dimpath);
  hobjreftalloc_init(&dimscales);
  do {
    context_t *ctx = get_context(ctxid);

    char *dimname = islongitude ? "XDim" : "YDim";
    if (make_dimpath(ctxid, gridname, 1, dimname, &dimpath) == -1) err_break(3)

    if(option == 2) {
      int i;
      for(i=start_ptr_file; i<end_ptr_file; i++) {
        if(strcmp(dimname, ctx->dim_data_name[i]) == 0)
          break;
      }
      if(i < end_ptr_file) {/* use file */
	pos = i;
        fill_xydim_scale_with_data(ctxid, dimname, coord->len, dimpath);
	goto FILLED;
      }
    } 

    atype = H5T_NATIVE_DOUBLE;
    if(strncmp("TES", ctx->instrument_name,ctx->len_of_instrument_name) == 0){
      if((fbuf = get_coord_from_geofield(ctxid, gridname, islongitude)) != NULL)
        atype = get_lonlat_type(ctxid, gridname, islongitude);
    }

    hsize_t dimsize = coord->len;

    if ((memtype = H5Tcopy(atype)) < 0) err1_break(3, "H5Tcopy")
    if ((spaceid = H5Screate_simple(1, &dimsize, NULL)) < 0) err1_break(3, "H5Screate_simple")

    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dsetid = H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0) err1_break(3, "H5Dopen")
    if (fbuf == NULL) { 
      if (H5Dwrite(dsetid, memtype, spaceid, spaceid, H5P_DEFAULT, coord->s) < 0) err1_break(3, "H5Dwrite")
    } else {
      if (H5Dwrite(dsetid, memtype, spaceid, spaceid, H5P_DEFAULT, fbuf) < 0) err1_break(3, "H5Dwrite")
    }
FILLED:
    if (!hobjreftalloc_ready(&dimscales, 1)) nomem1_break("obj ref")
    if (H5Rcreate(&dimscales.s[0], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err1_break(3, "H5Rcreate")
    dimscales.len++;

    if (attach_dimension(ctxid, dimpath.s, &dimscales) == -1) err_break(3)

    if(!g_quiet)
      msg2("fill 1d dimension", dimpath.s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (spaceid != -1) H5Sclose(spaceid);
    if (memtype != -1) H5Tclose(memtype);
    if (dsetid != -1) H5Dclose(dsetid);
    if (atype != -1) H5Tclose(atype);
    hobjreftalloc_free(&dimscales);
    challoc_free(&dimpath);
    if(fbuf != NULL) free(fbuf);
  }

  return succeeded ? 0 : -1;
}

int fill_hdf5_2d_dimension(hid_t ctxid, const char *gridname, int islongitude, const doublealloc *coord, int ydimmajor, long xsize, long ysize)
{
  hid_t dsetid = -1;
  hid_t memtype = -1;
  hid_t dsettype = -1;
  hid_t spaceid = -1;
  challoc cvarpath;
  challoc dimpath;
  hobjreftalloc dimscales;
  int succeeded = 0;

  challoc_init(&cvarpath);
  challoc_init(&dimpath);
  hobjreftalloc_init(&dimscales);
  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);
    const char *cvname = islongitude ? "lon" : "lat";
    hsize_t dimsizes[2];

    if (ydimmajor) { dimsizes[0] = ysize; dimsizes[1] = xsize; }
    else { dimsizes[0] = xsize; dimsizes[1] = ysize; }

    if ((memtype = H5Tcopy(H5T_NATIVE_DOUBLE)) < 0) err1_break(3, "H5Tcopy")
    if ((dsettype = H5Tcopy(H5T_IEEE_F32LE)) < 0) err1_break(3, "H5Tcopy")
    if ((spaceid = H5Screate_simple(2, dimsizes, NULL)) < 0) err1_break(3, "H5Screate_simple")

    if (make_dimpath(ctxid, gridname, 1, cvname, &cvarpath) == -1) err_break(3)
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dsetid = H5Dcreate(ctx->hdf5fd, cvarpath.s, dsettype, spaceid, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0) err2_break(3, "H5Dcreate", cvarpath.s)
    if (H5Dwrite(dsetid, memtype, spaceid, spaceid, H5P_DEFAULT, coord->s) < 0) err1_break(3, "H5Dwrite")

    if (!hobjreftalloc_ready(&dimscales, 2)) nomem1_break("obj ref")
    for (i = 0; i < 2; ++i) {
      const char *dimname = ydimmajor ? (i == 0 ? "YDim" : "XDim") : (i == 0 ? "XDim": "YDim");
      if (make_dimpath(ctxid, gridname, 1, dimname, &dimpath)) err_flagbreak(3, failinside)
      if (H5Rcreate(&dimscales.s[i], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err1_flagbreak(3, failinside, "H5Rcreate")
      dimscales.len++;
    }
    if (failinside) break;

    if (attach_dimension(ctxid, cvarpath.s, &dimscales) == -1) err_break(3)

      if(!g_quiet)
	msg2("fill 2d dimension", cvarpath.s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (spaceid != -1) H5Sclose(spaceid);
    if (dsettype != -1) H5Tclose(dsettype);
    if (memtype != -1) H5Tclose(memtype);
    if (dsetid != -1) H5Dclose(dsetid);
    hobjreftalloc_free(&dimscales);
    challoc_free(&dimpath);
    challoc_free(&cvarpath);
  }
  return succeeded ? 0 : -1;
}

static int write_attributes(hid_t ctxid, const char *gridswathname, int isgrid, int isdim, const char *dimname, int numattr, const char **attrnames, const char **attrvalues)
{
  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;

  challoc_init(&dimpath);
  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);

    if (isdim) {
      if (make_dimpath(ctxid, gridswathname, isgrid, dimname, &dimpath) == -1) err_break(3)
    }
    else {
      if (make_fieldpath(ctxid, gridswathname, isgrid, dimname, 0, &dimpath) == -1) err_break(3)
    }
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dsetid = H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0) err1_break(3, "H5Dopen")

    for (i = 0; i < numattr; ++i) {
      if (write_attribute_string(ctxid, dsetid, attrnames[i], attrvalues[i]) == -1) err2_flagbreak(3, failinside, "attribute", attrnames[i])
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
    challoc_free(&dimpath);
  }
  return succeeded ? 0 : -1;
}

int conform_cf_conventions_grid_1d(hid_t ctxid, const char *gridname, int islongitude)
{
  int succeeded = 0;

  do {
    const char *dimname = islongitude ? "XDim" : "YDim";
    const char *longname = islongitude ? "longitude" : "latitude";
    const char *unit = islongitude ? "degrees_east" : "degrees_north";
    const char *attrnames[2] = { "long_name", "units" };
    const char *attrvalues[2] = { longname, unit };

    if (write_attributes(ctxid, gridname, 1, 1, dimname, 2, attrnames, attrvalues) == -1) err_break(3)

    msg2("conform to CF convention 1.5 section 5.1", dimname);

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

int conform_cf_conventions_grid_2d_dim(hid_t ctxid, const char *gridname, int islongitude)
{
  int succeeded = 0;

  do {
    /* Only "axis" and "long_name' will be written. This is slightly different from
     * the CF Conventions in that "units" attribute is missing.
     */
    const char *dimname = islongitude ? "XDim" : "YDim";
    const char *axis = islongitude ? "X" : "Y";
    const char *longname = islongitude ? "x-coordinate in Cartesian system" : "y-coordinate in Cartesian system";
    const char *attrnames[2] = { "axis", "long_name" };
    const char *attrvalues[2] = { axis, longname };

    if (write_attributes(ctxid, gridname, 1, 1, dimname, 2, attrnames, attrvalues) == -1) err_break(3)

    msg2("conform to CF convention 1.5 section 5.2", dimname);

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

int conform_cf_conventions_grid_2d_coordvar(hid_t ctxid, const char *gridname, int islongitude, challoc *coordinates)
{
  int succeeded = 0;

  do {
    /* Both "lon" and "lat" are two-dimensional coordinate variables that
     * this tool generates to handle two-dimensional latitude and longitude case
     * such as sinusoidal projection, polar stereographic projection, etc.
     */
    const char *cvname = islongitude ? "lon" : "lat";
    const char *longname = islongitude ? "longitude" : "latitude";
    const char *unit = islongitude ? "degrees_east" : "degrees_north";
    const char *attrnames[2] = { "long_name", "units" };
    const char *attrvalues[2] = { longname, unit };

    if (write_attributes(ctxid, gridname, 1, 0, cvname, 2, attrnames, attrvalues) == -1) err_break(3)
    if (coordinates->len > 0) {
      if (!challoc_cats(coordinates, " ")) nomem1_break("coordinates")
    }
    if (!challoc_cats(coordinates, cvname)) nomem1_break("coordinates")

    msg2("conform to CF convention 1.5 section 5.2", cvname);

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

int conform_cf_conventions_2d_data(hid_t ctxid, const char *gridswathname, int isgrid, const fieldinfo_t *field, const challoc *coordinates)
{
  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;

  challoc_init(&dimpath);
  do {
    context_t *ctx = get_context(ctxid);

    if (make_fieldpath(ctxid, gridswathname, isgrid, field->name.s, 1, &dimpath) == -1) err_break(3)
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dsetid = H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0) err1_break(3, "H5Dopen")

    if (write_attribute_string(ctxid, dsetid, "coordinates", coordinates->s) == -1) err_break(3)

    msg2("conform to CF convention 1.5 section 5.2 (coordinates)", dimpath.s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
    challoc_free(&dimpath);
  }
  return succeeded ? 0 : -1;
}

herr_t fill_hdf5_dimension_on_attr(hid_t src, const char *attrname, const H5A_info_t *ainfo, void *arg)
{
  hid_t aid, tid, aid2, sid;
  hid_t dest = *(hid_t*)arg;
  hid_t retval;
  size_t size;
  
  /* We copy all atttibutes except "Title" and "DIMENSION_LIST" */
  if(strcmp(attrname, "Title")==0)
    return 0;
  /* else if(strcmp(attrname, "DIMENSION_LIST")==0) */
  /*  return 0; */

  if((aid = H5Aopen(src, attrname, H5P_DEFAULT)) < 0)
    return -1;
  if((tid = H5Aget_type(aid))<0)
    return -1;
  if((sid = H5Aget_space(aid))<0)
    return -1;
  if((size = H5Aget_storage_size(aid))<0)
    return -1;

  void *buf = malloc(size);
  if(buf==NULL) 
    return -1;
  if((retval = H5Aread(aid, tid, buf)) < 0)
    return -1;

  if((aid2=H5Acreate2(dest, attrname, tid, sid, H5P_DEFAULT, H5P_DEFAULT))<0)
    return -1;
  if((retval = H5Awrite(aid2, tid, buf))<0)
    return -1;
  return 0;
}

int fill_hdf5_swath_dimension(hid_t ctxid, const char *gridswathname, const dimension_t *dim)
{
  challoc dimpath;
  int succeeded = 0;
  hid_t source=-1, tid=-1, spaceid=-1, dest=-1;
  hssize_t npoints;
  size_t elemsize;
  void *buf = NULL;

  challoc_init(&dimpath);
  do {
    int i;

    context_t *ctx = get_context(ctxid);
    
    hsize_t dimsize = dim->dimsize;
    hsize_t maxdimsize = dim->dimsize;

    if(dimsize < 0)
      return -1;

    if (dimsize == H5S_UNLIMITED)
      dimsize = 0;

    if (make_dimpath(ctxid, gridswathname, 0, dim->name.s, &dimpath) == -1) err_break(3);
      
    switch(option) {
      case 0:
	{
	  put_att(ctxid, dim, dimpath);
	  break;
	}
      case 1:
	{
	  fill_dim_scale_with_index(ctxid, dim, dimpath);
	  break;
	}
      case 2:
	{
	  for(i=start_ptr_field; i<end_ptr_field; i++)
	  {
	    if(strcmp(dim->name.s, ctx->dimnames[i])==0)
	      break;
	  }
	  if(i < end_ptr_field) /* [dim, field] */
	  {
	    fill_dim_scale_with_field(ctxid, gridswathname, dim, dimpath, 0, i);
	    goto DONE;
	  } 

	  for(i=start_ptr_file; i<end_ptr_file; i++) {
	    if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
	      break;
	  }
	  if(i < end_ptr_file) {/* [dim, file] */
	    pos = i;
	    fill_dim_scale_with_data(ctxid, dim, dimpath);
	    goto DONE;
	  }

	  for(i=start_ptr_index; i<end_ptr_index; i++) {
            if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
	       break;
	  }
	  if(i < end_ptr_index) /* [dim] */
	    fill_dim_scale_with_index(ctxid, dim, dimpath);
	}/* end of case 2 */	  
    }/* end of switch */
DONE:
    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&dimpath);
    if(buf) free(buf);
    if (spaceid != -1) H5Sclose(spaceid);
    if (tid != -1) H5Tclose(tid);
    if (dest != -1) H5Dclose(dest);
    if (source != -1) H5Dclose(source);
  }
 
  return succeeded ? 0 : -1;
}

int put_att(hid_t ctxid, const dimension_t *dim, const challoc dimpath) {
  int succeeded=0;
  hid_t dest=-1;
  context_t *ctx;

  do {
    ctx = get_context(ctxid);
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dest= H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0)  err2_break(3, "open dataset", dimpath.s)
    if(H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(3, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(3);
    if(H5Dclose(dest) < 0) err2_break(3, "close dataset", dimpath.s);
    
    succeeded = 1;
    } while(0);

  return succeeded? 0: -1;
}

int fill_dim_scale_with_index(hid_t ctxid, const dimension_t *dim, const challoc dimpath) {
  int succeeded=0;
  hid_t dest=-1;
  context_t *ctx;

  do {
    int i;
    int ds = dim->dimsize;

    ctx = get_context(ctxid);
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dest= H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0)  err2_break(3, "open dataset", dimpath.s)
    if(H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(3, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(3);
 
    int* padding_data =  malloc(ds*sizeof(int));
    for(i=0; i<ds; i++)
      padding_data[i] = i+1;
    if (H5Dwrite(dest, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, padding_data) < 0) err2_break(3, "write", dimpath.s)
    
    free(padding_data);
    if(H5Dclose(dest) < 0) err2_break(3, "close dataset", dimpath.s);
    
    succeeded = 1;
  } while(0);

  return succeeded? 0: -1;
}

int fill_xydim_scale_with_data(hid_t ctxid, const char *dimname, const int dimsize, const challoc dimpath) {
 
  hid_t dest=-1;
  context_t *ctx;
 
  int succeeded=0;

  do {
    ctx = get_context(ctxid);
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dest= H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0) err2_break(3, "open dataset", dimpath.s)
   
    if(H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(3, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dimname) == -1) err_break(3);
   
    float* coord = malloc(dimsize*sizeof(float));
    doublealloc da = ctx->dim_data[pos];
    if(dimsize != da.len) {
      printf("The required data length [%i] is different from the length [%i] of data file. Please check the data file to eliminate the inconsistency.\n", dimsize, da.len);
      exit(0);
    }
    int k;
    for(k=0; k<dimsize; k++) {
      coord[k] = (float)da.s[k];
    }
    if (H5Dwrite(dest, H5T_IEEE_F32LE, H5S_ALL, H5S_ALL, H5P_DEFAULT, coord) < 0) err2_break(3, "write", dimpath.s)
    free(coord);

    if(H5Dclose(dest) < 0) err2_break(3, "close dataset", dimpath.s);
    
    succeeded = 1;
  } while(0);

  return succeeded? 0: -1;
}

int fill_dim_scale_with_data(hid_t ctxid, const dimension_t *dim, const challoc dimpath) {
  
  hid_t dest=-1;
  context_t *ctx;

  int succeeded = 0;

  do {
    int ds = dim->dimsize;
    ctx = get_context(ctxid);
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dest= H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0) err2_break(3, "open dataset", dimpath.s)
    if (H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(3, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(3);

    float* coord = malloc(ds*sizeof(float));
    doublealloc da = ctx->dim_data[pos];
    if(ds != da.len) {
      printf("The required data length [%i] is different from the length [%i] of data file. Please check the data file to eliminate the inconsistency.\n", ds, da.len);
      exit(0); 
    }
    int k;
    for(k=0; k<ds; k++) {  
      coord[k] = (float)da.s[k];
    }
    if (H5Dwrite(dest, H5T_IEEE_F32LE, H5S_ALL, H5S_ALL, H5P_DEFAULT, coord) < 0) err2_break(3, "write", dimpath.s)
    free(coord);

    if(H5Dclose(dest) < 0) err2_break(3, "close dataset", dimpath.s);
    
    succeeded = 1;
  } while(0);

  return succeeded? 0: -1;
}

int fill_dim_scale_with_field(hid_t ctxid, const char *gridswathname, const dimension_t *dim, const challoc dimpath, int isgrid, const int i) {
  hid_t source=-1, tid=-1, spaceid=-1, dest=-1;
  hssize_t npoints;
  size_t elemsize;
  void *buf = NULL;

  char geopath[1024];
  int succeeded = 0;
      
  do {
    context_t *ctx = get_context(ctxid);
    const char *geoname = ctx->geonames[i];

    hsize_t dimsize = dim->dimsize;
    hsize_t maxdimsize = dim->dimsize;

    if(dimsize == H5S_UNLIMITED)  
      dimsize = 0;

    if(isgrid)
      sprintf(geopath, EOSDATAFIELD, gridswathname, geoname);
    else
      sprintf(geopath, (ctx->za==0)? EOSGEOFIELD: EOSGEOFIELD2,  gridswathname, geoname);
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((source = H5Dopen(ctx->hdf5fd, geopath, H5P_DEFAULT)) < 0) 
    {
      fprintf(stderr, "Warning: Dimension name %s cannot be filled since %s does not exists.\n", dim->name.s, geopath);
	break;
    }

    if ((tid = H5Dget_type(source)) < 0) err2_break(3, "get type", geopath)
    if ((spaceid = H5Dget_space(source)) < 0) err2_break(3, "get space", geopath)

    if ((npoints = H5Sget_simple_extent_npoints(spaceid)) < 0) err2_break(3, "get npoints", geopath)
    if ((elemsize = H5Tget_size(tid)) == 0) err2_break(3, "get size", geopath)
    if ((buf = malloc(npoints * elemsize)) == NULL) nomem2_break("clone", geopath)
    if (H5Dread(source, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(3, "read", geopath)
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dest= H5Dopen(ctx->hdf5fd, dimpath.s, H5P_DEFAULT)) < 0) err2_break(3, "open dataset", dimpath.s);
    if (H5Dwrite(dest, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(3, "write", dimpath.s)

    /* Rename NAME attribute
       The original value of the NAME attribute is 
       "This is a netCDF dimension but not a netCDF variable".
       With this value, NetCDF cannot read any value in it. */
    if(H5Aexists(dest, "NAME") > 0) 
      if(H5Adelete(dest, "NAME")<0) err2_break(3, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(3);

    /* Copy attributes from the original dimension scale */
     hsize_t n=0;
     if(H5Aiterate2(source, H5_INDEX_CRT_ORDER, H5_ITER_NATIVE, &n, fill_hdf5_dimension_on_attr, (void*)&dest)<0)
       err2_break(3, "copy attributes", dimpath.s)

    /* Remove DIMENSION_LIST */
    if(H5Adelete(dest, "DIMENSION_LIST")<0) err2_break(3, "del dimension list", dimpath.s);
    succeeded = 1;
  } while(0);

  if(1) {
    if(buf) free(buf);
    if(spaceid != -1) H5Sclose(spaceid);
    if(tid != -1) H5Tclose(tid);
    if(dest != -1) H5Dclose(dest);
    if(source != -1) H5Dclose(source);
  }

  return succeeded? 0: -1;
}
// vim:ts=8:sw=2:sts=2:cindent
